import lunr
from sabledocs.proto_model import Enum, Message, Package, SableContext, Service


def build_search_index(sable_context: SableContext) -> tuple[dict[str, dict[str, str]], lunr.index.Index]:
    documents = []

    for package in sable_context.non_hidden_packages:
        doc = {
            "package": package.name,
            "url": f"{package.name}.html",
            "content": build_package_content(package)
        }

        documents.append(doc)

    builder = lunr.get_default_builder()
    #builder.metadata_whitelist.append("position") This can be enabled to get position information in the search results, not used yet.

    idx = lunr.lunr(
        ref="package", fields=("content", "content", "url"), documents=documents, builder=builder
    )

    documents_dict = {d["package"]:d for d in documents}

    return (documents_dict, idx)


def build_package_content(package: Package):
    acc = ""
    acc += package.name + "\n"
    acc += package.description + "\n"
    for service in package.services:
        acc = append_service_content(acc, service)

    for message in package.messages:
        acc = append_message_content(acc, message)

    for enum in package.enums:
        acc = append_enum_content(acc, enum)
    
    return acc


def append_service_content(acc: str, service: Service):
    for name_segment in service.full_name.split("."):
        acc += name_segment + "\n"

    acc += service.description + "\n"
    for method in service.methods:
        acc += method.name + "\n"
        acc += method.description + "\n"
        for name_segment in method.request.full_type.split("."):
            acc += name_segment + "\n"

        for name_segment in method.response.full_type.split("."):
            acc += name_segment + "\n"
    
    return acc


def append_message_content(acc: str, message: Message):
    for name_segment in message.full_name.split("."):
        acc += name_segment + "\n"

    for field in message.fields:
        acc += field.name + "\n"
        acc += field.description + "\n"
    
    return acc


def append_enum_content(acc: str, enum: Enum):
    for name_segment in enum.full_name.split("."):
        acc += name_segment + "\n"

    acc += enum.description + "\n"
    for value in enum.values:
        acc += value.name + "\n"
        acc += value.description + "\n"
    
    return acc



from pprint import pformat
from typing import Optional
from sabledocs.sable_config import SableConfig


class CodeItem:
    def __init__(self):
        self.name = ''
        self.description = ''
        self.description_html = ''
        self.source_file_path = ''
        self.line_number = 0
        self.repository_url: Optional[str] = ''


class MessageField(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.number = 0
        self.label = ''
        self.type = ''
        self.full_type = ''
        self.default_value = ''
        self.package: Optional[Package] = None
        self.is_package_hidden = False
        self.type_kind = "UNKNOWN"
        self.oneof_name: Optional[str] = None

    def __repr__(self):
        filtered_vars = dict(filter(lambda elem: elem[0] != "package", vars(self).items()))
        return pformat(filtered_vars, indent=4, width=1)


class OneOfFieldGroup:
    def __init__(self, name: str, fields: list[MessageField]):
        self.name = name
        self.fields: list[MessageField] = fields


class Message(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.full_name = ''
        self.is_map_entry = False
        self.fields: list[MessageField] = []
        self.oneof_field_groups: list[OneOfFieldGroup] = []
        self.parent_message = None
        self.package: Optional[Package] = None
        self.type_kind = "MESSAGE"

    @property
    def full_type(self):
        return self.full_name
    
    @property
    def has_any_fields_with_default_value(self):
        return any(filter(lambda f: f.default_value, self.fields))

    @property
    def non_oneof_fields(self):
        return [f for f in self.fields if not f.oneof_name]

    def __repr__(self):
        return pformat(vars(self), indent=4, width=1)


class EnumValue(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.number = 0


class Enum(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.full_name = ''
        self.values: list[EnumValue] = []
        self.parent_message = None
        self.package: Optional[Package] = None
        self.type_kind = "ENUM"

    def __repr__(self):
        return pformat(vars(self), indent=4, width=1)


class ServiceMethodArgument(CodeItem):
    def __init__(self, type: str, full_type: str, type_kind: str):
        CodeItem.__init__(self)
        self.type = type
        self.full_type = full_type
        self.type_kind = type_kind
        self.package: Optional[Package] = None


class ServiceMethod(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.request = ServiceMethodArgument("", "", "")
        self.response = ServiceMethodArgument("", "", "")


class Service(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.full_name = ''
        self.methods: list[ServiceMethod] = []


class Package(CodeItem):
    def __init__(self):
        CodeItem.__init__(self)
        self.messages = []
        self.enums = []
        self.services = []

    def __repr__(self):
        return pformat(vars(self), indent=4, width=1)


class LocationInfo:
    def __init__(self, line_number, comments):
        self.line_number = line_number
        self.comments = comments


class SableContext:
    def __init__(self, packages: list[Package], all_messages: list[Message], all_enums: list[Enum], sable_config: SableConfig):
        self.packages = packages
        self.all_messages = all_messages
        self.all_enums = all_enums
        self.sable_config = sable_config

    @property
    def non_hidden_packages(self):
        return [p for p in self.packages if p.name not in self.sable_config.hidden_packages]




import json
import markdown
import os
import pprint
import sys
from shutil import copy, copytree, rmtree

from jinja2 import Environment, FileSystemLoader, select_autoescape

from sabledocs.lunr_search import build_search_index
from sabledocs.proto_descriptor_parser import parse_proto_descriptor
from sabledocs.sable_config import SableConfig


def check_python_version():
    print(f"Python {sys.version}")

    if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 11):
        print("WARNING: Sabledocs requires Python version 3.11 or higher")


def cli():
    print("Starting Sabledocs")
    check_python_version()

    print()
    print("Starting documentation generation.")
    sable_config = SableConfig("sabledocs.toml")

    if not os.path.exists(sable_config.input_descriptor_file):
        print()
        print(f'ERROR: The Proto descriptor file {sable_config.input_descriptor_file} does not exist. Exiting.')
        return

    # Execute the main processing of the Proto contracts.
    sable_context = parse_proto_descriptor(sable_config)

    if sable_config.template != "_default":
        print()
        print('WARNING: The "template" config parameter is deprecated, it will be removed in a future version. The field template-path should be used instead.')
        template_base_dir = f"templates/{sable_config.template}"
    else:
        template_base_dir = sable_config.template_path if sable_config.template_path else os.path.join(os.path.dirname(__file__), "templates", "_default")

    jinja_env = Environment(
        loader=FileSystemLoader(searchpath=template_base_dir),
        autoescape=select_autoescape()
    )

    package_template = jinja_env.get_template("package.html")

    if not os.path.exists(sable_config.output_dir):
        os.makedirs(sable_config.output_dir)

    # NOTE: When the output files are generated, the encode('utf-8') option has to be used, otherwise Unicode characters like © can end up garbled.
    for package in sable_context.non_hidden_packages:
        with open(os.path.join(sable_config.output_dir, f'{package.name if package.name else "__default"}.html'), 'wb') as fh:
            output = package_template.render(
                sable_config=sable_config,
                package=package,
                packages=sable_context.packages,
                non_hidden_packages=sable_context.non_hidden_packages,
                all_messages=sable_context.all_messages,
                all_enums=sable_context.all_enums).encode('utf-8')

            fh.write(output)

    main_page_content = ""

    if sable_config.main_page_content_file != "":
        print()
        if os.path.exists(sable_config.main_page_content_file):
            print(f"Found main content page, {sable_config.main_page_content_file}.")
            with open(sable_config.main_page_content_file, mode='r') as main_page_content_file:
                main_page_content = markdown.markdown(main_page_content_file.read(), extensions=sable_config.markdown_extensions)
        else:
            print(f"WARNING: The configured main content page, {sable_config.main_page_content_file} was not found.")

    with open(os.path.join(sable_config.output_dir, 'index.html'), 'wb') as fh:
        output = jinja_env.get_template("index.html").render(
            sable_config = sable_config,
            main_page_content = main_page_content,
            packages = sable_context.packages,
            non_hidden_packages=sable_context.non_hidden_packages,
            all_messages = sable_context.all_messages,
            all_enums = sable_context.all_enums).encode('utf-8')

        fh.write(output)

    if sable_config.enable_lunr_search:
        (search_documents, search_index) = build_search_index(sable_context)

        with open(os.path.join(sable_config.output_dir, 'search.html'), 'wb') as fh:
            output = jinja_env.get_template("search.html").render(
                sable_config = sable_config,
                search_documents = json.dumps(search_documents),
                search_index = json.dumps(search_index.serialize())).encode('utf-8')

            fh.write(output)

    index_abs_path = os.path.abspath(os.path.join(sable_config.output_dir, "index.html"))

    output_static_path = os.path.join(sable_config.output_dir, "static")

    if os.path.exists(output_static_path):
        # This is needed, because shutils.copytree cannot copy to a target folder which already exists.
        rmtree(output_static_path)

    copytree(os.path.join(template_base_dir, "static"), output_static_path)

    if os.path.exists("static") and os.path.isdir("static"):
        print("Copying static content from the folder 'static'.")
        for root, _, files in os.walk("static"):
            dir_path = "" if root == "static" else root.removeprefix("static\\")
            dest_dir_path = os.path.join(sable_config.output_dir, dir_path)
            if dir_path != "":
                if not os.path.exists(dest_dir_path):
                    os.makedirs(dest_dir_path)

            for f in files:
                src_file_path = os.path.join(root, f)
                dest_file_path = os.path.join(dest_dir_path, f)
                if not os.path.exists(dest_file_path):
                    copy(src_file_path, dest_file_path)

    if sable_config.extra_template_path != "":
        print(f"Rendering extra Jinja templates from, {sable_config.extra_template_path}")
        jinja_extra_env = Environment(
            loader=FileSystemLoader(searchpath=sable_config.extra_template_path),
            autoescape=select_autoescape()
        )
        for root, _, files in os.walk(sable_config.extra_template_path):
            dir_path = "" if root == sable_config.extra_template_path else (
                root.removeprefix(sable_config.extra_template_path).rstrip("/\\"))
            if "/_" in dir_path or "\\_" in dir_path:  # ignore subdirectories that start with "_"
                continue
            for file in files:
                if not file.endswith(sable_config.extra_template_suffix):
                    continue
                file_path = file if dir_path == "" else str(os.path.join(dir_path, file))
                print(f"Rendering extra Jinja template, {file_path}")
                with open(os.path.join(sable_config.output_dir, file_path), 'wb') as fh:
                    output = jinja_extra_env.get_template(file_path).render(
                        sable_config=sable_config).encode('utf-8')

                    fh.write(output)
    print()
    print(f"Building documentation done. It can be opened with {index_abs_path}")

if __name__ == '__main__':  # pragma: no cover
    cli()
